/* ============================================================
 * Author: Aditya Nagrath
 * Date  : 2008-01-24
 * Copyright 2008 Palm, Inc. All rights reserved.
 * ============================================================ */
#include <glib.h>

#include <PmIpcServer.h>
#include "DocumentServer.h"
#include "DocCommander.h"
#include <stdio.h>



static DocumentServer* dpInstance     = 0;
static bool           gInit          = false;
static PmIpcServer_t* dpServer       = 0;

static bool
PrvClientConnect(PmIpcClientContext_t* pClientCtxt, bool attached, void* pUserData);

static void
PrvCmdOpen(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	   const PmIpcVariant_t* pArg1,  const PmIpcVariant_t* pArg2);

static void
PrvCmdInitialize(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand);

static void
PrvCmdSetPassword(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	   const PmIpcVariant_t* pArg1);

static void
PrvCmdRegisterBuffer(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
                     const PmIpcVariant_t* pArg1, const PmIpcVariant_t* pArg2);

static void
PrvCmdRender(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	     const PmIpcVariant_t* pArg1, const PmIpcVariant_t* pArg2, const PmIpcVariant_t* pArg3,
	     const PmIpcVariant_t* pArg4, const PmIpcVariant_t* pArg5);

static void
PrvCmdNextPage(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand);

static void
PrvCmdPreviousPage(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand);

static void
PrvCmdJumpToPage(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	   const PmIpcVariant_t* pArg1);

static void
PrvCmdSetZoom(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	   const PmIpcVariant_t* pArg1);

static void
PrvCmdGetZoom(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand);


static void
PrvCmdSetRotation(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	   const PmIpcVariant_t* pArg1);

static void
PrvCmdGetRotation(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand);


static void
PrvCmdGetCurrentPageNumber(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand);

static void
PrvCmdGetPageCount(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand);

static void
PrvCmdGetPageDimensions(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand);

static void
PrvCmdHasLinkAtLocation(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
			const PmIpcVariant_t* pArg1, const PmIpcVariant_t* pArg2,
			const PmIpcVariant_t* pArg3);

static void
PrvCmdOutlineRoot(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		  const char* pCommand);

static void
PrvCmdGetOutlineNodeText(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand);


static void
PrvCmdOutlineGotoChild(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand);

static void
PrvCmdOutlineHasChild(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand);

static void
PrvCmdOutlineGotoNextSibling(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand);

static void
PrvCmdOutlineHasNextSibling(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand);

			
static void
PrvCmdGetOutlineLinkPageNumber(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand);

static void
PrvCmdCaseInsensitiveTextSearch(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
			       const char* pCommand,const PmIpcVariant_t* pArg1,
			       const PmIpcVariant_t* pArg2, const PmIpcVariant_t* pArg3);

//funcs that probably wont be needed for anything besides test
static void
PrvCmdDumpPageText(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand);

static void
PrvCmdTranslateScreenPoint(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
			   const char* pCommand, const PmIpcVariant_t* pArg1,
			   const PmIpcVariant_t* pArg2);

static void
PrvCmdTranslateScreenDelta(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
			   const char* pCommand, const PmIpcVariant_t* pArg1,
			   const PmIpcVariant_t* pArg2);

DocumentServer::DocumentServer()
{
    dpInstance = this;
}

DocumentServer::~DocumentServer()
{
    dpInstance = 0;
}

DocumentServer*
DocumentServer::instance()
{
    if (!dpInstance) {
      new DocumentServer;
    }

    return dpInstance;
}

bool
DocumentServer::init()
{
    PmErr      err = kPmErrNone;
    GMainLoop* pLoop;
    void*      pTmp;
    
    if (gInit) {
        return true;
    }

    err = PmIpcServerCreate(&dpServer);
    if (err != kPmErrNone) {
        BERR("Failed to create up server: 0x%08x", err);
        goto Failed;
    }

    err = PmIpcServerBringUp(dpServer, "document", PrvClientConnect, this);
    if (err != kPmErrNone) {
        BERR("Failed to create up bringup server: 0x%08x", err);
        goto Failed;
    }

    err = PmIpcServerGetMainLoop(dpServer, &pTmp);
    if (err != kPmErrNone) {
        BERR("Failed to get main loop from server: 0x%08x", err);
        goto Failed;
    }
    pLoop = (GMainLoop*) pTmp;


    
    err = PmIpcServerRegisterHandler(dpServer, "d.Open",
                                     (PmIpcServerCommandCallback) PrvCmdOpen,
                                     this,
                                     kPmIpcTypeString, // virtual filename
                                     kPmIpcTypeUint32, // virtual ID
                                     kPmIpcTypeNull);
    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }

    err = PmIpcServerRegisterHandler(dpServer, "d.Initialize",
                                     (PmIpcServerCommandCallback) PrvCmdInitialize,
                                     this,
                                     kPmIpcTypeNull);
    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }

    err = PmIpcServerRegisterHandler(dpServer, "d.SetPassword",
                                     (PmIpcServerCommandCallback) PrvCmdSetPassword,
                                     this,
                                     kPmIpcTypeString, // password
                                     kPmIpcTypeNull);
    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }
    
    err = PmIpcServerRegisterHandler(dpServer, "d.RegisterBuffer",
                                     (PmIpcServerCommandCallback) PrvCmdRegisterBuffer,
                                     this,
                                     kPmIpcTypeUint32, // buffer id
                                     kPmIpcTypeUint32, // buffer size
                                     kPmIpcTypeNull);
    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }
    err = PmIpcServerRegisterHandler(dpServer, "d.HasLinkAtLocation",
                                     (PmIpcServerCommandCallback) PrvCmdHasLinkAtLocation,
                                     this,
                                     kPmIpcTypeUint32, // int x
                                     kPmIpcTypeUint32, // int y
                                     kPmIpcTypeBool, // Return the link!
                                     kPmIpcTypeNull);
    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }

    err = PmIpcServerRegisterHandler(dpServer, "d.Render",
                                     (PmIpcServerCommandCallback) PrvCmdRender,
                                     this,
                                     kPmIpcTypeUint32, // int center x
                                     kPmIpcTypeUint32, // int center y
                                     kPmIpcTypeUint32, // int width
                                     kPmIpcTypeUint32, // int height
                                     kPmIpcTypeUint32, // int Render to buffer with shmid
                                     kPmIpcTypeNull);
    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }

    err = PmIpcServerRegisterHandler(dpServer, "d.NextPage",
                                     (PmIpcServerCommandCallback) PrvCmdNextPage,
                                     this,
                                     kPmIpcTypeNull);
    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }

    err = PmIpcServerRegisterHandler(dpServer, "d.PreviousPage",
                                     (PmIpcServerCommandCallback) PrvCmdPreviousPage,
                                     this,
                                     kPmIpcTypeNull);
    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }
  
    err = PmIpcServerRegisterHandler(dpServer, "d.JumpToPage",
                                     (PmIpcServerCommandCallback) PrvCmdJumpToPage,
                                     this,
                                     kPmIpcTypeUint32, // Page Number
                                     kPmIpcTypeNull);

    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }

    err = PmIpcServerRegisterHandler(dpServer, "d.SetZoom",
                                     (PmIpcServerCommandCallback) PrvCmdSetZoom,
                                     this,
                                     kPmIpcTypeDouble, // Zoom Level
                                     kPmIpcTypeNull);

    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }


    err = PmIpcServerRegisterHandler(dpServer, "d.GetZoom",
                                     (PmIpcServerCommandCallback) PrvCmdGetZoom,
                                     this,
				     kPmIpcTypeNull);

    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }

    err = PmIpcServerRegisterHandler(dpServer, "d.SetRotation",
                                     (PmIpcServerCommandCallback) PrvCmdSetRotation,
                                     this,
                                     kPmIpcTypeInt32, // Rotation in degrees
                                     kPmIpcTypeNull);

    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }

    err = PmIpcServerRegisterHandler(dpServer, "d.GetRotation",
                                     (PmIpcServerCommandCallback) PrvCmdGetRotation,
                                     this,
                                     kPmIpcTypeNull);

    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }

    err |= PmIpcServerRegisterHandler(dpServer, "d.GetCurrentPageNumber",
                                     (PmIpcServerCommandCallback) PrvCmdGetCurrentPageNumber,
                                     this,
                                     kPmIpcTypeNull);

    err |= PmIpcServerRegisterHandler(dpServer, "d.GetPageCount",
				     (PmIpcServerCommandCallback) PrvCmdGetPageCount,
				     this,
				     kPmIpcTypeNull);

    err |= PmIpcServerRegisterHandler(dpServer, "d.GetPageDimensions",
				     (PmIpcServerCommandCallback) PrvCmdGetPageDimensions,
				     this,
				     kPmIpcTypeNull);

    err |= PmIpcServerRegisterHandler(dpServer, "d.OutlineRoot",
				     (PmIpcServerCommandCallback) PrvCmdOutlineRoot,
				     this,
				     kPmIpcTypeNull);

    err |= PmIpcServerRegisterHandler(dpServer, "d.GetOutlineNodeText",
				     (PmIpcServerCommandCallback) PrvCmdGetOutlineNodeText,
				     this,
				     kPmIpcTypeNull);

    err |= PmIpcServerRegisterHandler(dpServer, "d.OutlineGotoNextSibling",
				     (PmIpcServerCommandCallback) PrvCmdOutlineGotoNextSibling,
				     this,
				     kPmIpcTypeNull);

    err |= PmIpcServerRegisterHandler(dpServer, "d.OutlineGotoChild",
				     (PmIpcServerCommandCallback) PrvCmdOutlineGotoChild,
				     this,
				     kPmIpcTypeNull);

    err |= PmIpcServerRegisterHandler(dpServer, "d.OutlineHasNextSibling",
				     (PmIpcServerCommandCallback) PrvCmdOutlineHasNextSibling,
				     this,
				     kPmIpcTypeNull);

    err |= PmIpcServerRegisterHandler(dpServer, "d.OutlineHasChild",
				     (PmIpcServerCommandCallback) PrvCmdOutlineHasChild,
				     this,
				     kPmIpcTypeNull);

    err |= PmIpcServerRegisterHandler(dpServer, "d.GetOutlineLinkPageNumber",
				     (PmIpcServerCommandCallback) PrvCmdGetOutlineLinkPageNumber,
				     this,
				     kPmIpcTypeNull);

    err |= PmIpcServerRegisterHandler(dpServer, "d.CaseInsensitiveTextSearch",
                                     (PmIpcServerCommandCallback) PrvCmdCaseInsensitiveTextSearch,
                                     this,
                                     kPmIpcTypeString, // text
                                     kPmIpcTypeUint32, // int startpage
                                     kPmIpcTypeUint32, // int endpage
                                     kPmIpcTypeNull);


    err |= PmIpcServerRegisterHandler(dpServer, "d.DumpPageText",
				     (PmIpcServerCommandCallback) PrvCmdDumpPageText,
				     this,
				     kPmIpcTypeNull);


    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }

    err = PmIpcServerRegisterHandler(dpServer, "d.TranslateScreenPoint",
                                     (PmIpcServerCommandCallback) PrvCmdTranslateScreenPoint,
                                     this,
                                     kPmIpcTypeUint32, // int x
                                     kPmIpcTypeUint32, // int y
                                     kPmIpcTypeNull);
    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }



    err = PmIpcServerRegisterHandler(dpServer, "d.TranslateScreenDelta",
                                     (PmIpcServerCommandCallback) PrvCmdTranslateScreenDelta,
                                     this,
                                     kPmIpcTypeUint32, // int x
                                     kPmIpcTypeUint32, // int y
                                     kPmIpcTypeNull);
    if (err != kPmErrNone) {
        BERR("Failed to register handler: 0x%08x", err);
        goto Failed;
    }

	// ... Add more messages
	
    gInit = true;

    return true;

 Failed:

    if (dpServer) {
        PmIpcServerBringDown(dpServer);
        dpServer = 0;
    }

    return false;
}

bool
DocumentServer::run()
{
    PmErr err;
    
    if (!gInit) {
        return false;
    }

    err = PmIpcServerRunLoop(dpServer);
    if (err != kPmErrNone) {
        BERR("Palm IPC Error: 0x%08x", err);
        return false;
    }
    
    return true; 
}


bool
PrvClientConnect(PmIpcClientContext_t* pClientCtxt, bool attached, void* pUserData)
{
    //BDBG("Client %s : %p", attached ? "attached" : "detached", pClientCtxt);

    if (!attached) {
        DocCommander* pCom = 0;
        void*        pPtr  = 0;

        PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
        pCom = (DocCommander*) pPtr;
        if (pCom) {
            delete pCom;
            PmIpcServerSetPrivateData(pClientCtxt, 0);
        }
    }
    
    return true;
}

void
PrvCmdOpen(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	   const PmIpcVariant_t* pArg1,  const PmIpcVariant_t* pArg2)
{
  //BDBG("Got an Open Command!\n");
    PmErr        err   = kPmErrNone;
    DocCommander* pCom = 0;
    void*        pPtr  = 0;
    
    PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
    pCom = (DocCommander*) pPtr;
    if (pCom) {
        BERR("Client already sent Open command: %p", pClientCtxt);
        err = (uint32_t) -10;
        goto Done;
    }

    // XXX: ONce we get this file compiling.. we need to send the context to
    // to get responses back to the java side!!!
    pCom = new DocCommander((char *)PmIpcVariantGetString(pArg1), PmIpcVariantGetUint32(pArg2));
    err = -1*pCom->HasInitError();
    if (pCom== NULL) {
      BERR("Failed to allocated the Commander: %p %p", pClientCtxt, pCom);
        err = (uint32_t) -10;
        goto Done;
    }
 
    PmIpcServerSetPrivateData(pClientCtxt, pCom);

Done:
    //BDBG("Server returning %d\n", err);
    PmIpcServerSendSyncResponse(pClientCtxt,
                                kPmIpcTypeUint32, err,
                                kPmIpcTypeNull);
}

void
PrvCmdInitialize(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand)
{
  //BDBG("Got an Initialize Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  err = (uint32_t)pCom->Initialize();
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}

void
PrvCmdSetPassword(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	   const PmIpcVariant_t* pArg1)
{
  //BDBG("Got an SetPassword Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  
  pCom->SetPassword((char *)PmIpcVariantGetString(pArg1));
  if(pCom->NeedsPassword())
  {
    BERR("Password INCORRECT.  Notify client!");
    err = (uint32_t) -2;
    goto Done;
  }

 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}

void
PrvCmdRegisterBuffer(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
                     const PmIpcVariant_t* pArg1, const PmIpcVariant_t* pArg2)
{
  //BDBG("Got an RegisterBuffer Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  pCom->RegisterBuffer(PmIpcVariantGetUint32(pArg1),PmIpcVariantGetUint32(pArg2));
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}

static void
PrvCmdRender(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	     const PmIpcVariant_t* pArg1, const PmIpcVariant_t* pArg2, const PmIpcVariant_t* pArg3,
	     const PmIpcVariant_t* pArg4, const PmIpcVariant_t* pArg5)
{
  //BDBG("Got an RegisterBuffer Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  err = pCom->RenderToBuffer(PmIpcVariantGetUint32(pArg1),PmIpcVariantGetUint32(pArg2),
			     PmIpcVariantGetUint32(pArg3),PmIpcVariantGetUint32(pArg4),
			     PmIpcVariantGetUint32(pArg5));
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);

}

void
PrvCmdNextPage(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand)
{
  //BDBG("Got an NextPage Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  
  err =(uint32_t) pCom->NextPage();
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}

void
PrvCmdPreviousPage(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand)
{
  //BDBG("Got an PreviousPage Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  err =(uint32_t) pCom->PreviousPage();
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}

void
PrvCmdJumpToPage(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	   const PmIpcVariant_t* pArg1)
{
  //BDBG("Got an Open Command!\n");
   PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  err =(uint32_t) pCom->JumpToPage(PmIpcVariantGetUint32(pArg1));
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}

static void
PrvCmdSetZoom(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	   const PmIpcVariant_t* pArg1)
{
  //BDBG("Got an SetZoom Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  pCom->SetZoom(PmIpcVariantGetDouble(pArg1));
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}

static void
PrvCmdGetZoom(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand)
{
  //BDBG("Got an GetZoom Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  double       retval =0;

  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  retval = pCom->GetZoom();
 Done:
  //BDBG("Server returning %d, %e\n", err, retval);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeDouble, retval,
			      kPmIpcTypeNull);
}

static void
PrvCmdSetRotation(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
	   const PmIpcVariant_t* pArg1)
{
  //BDBG("Got an SetRotation Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }

  pCom->SetRotation(PmIpcVariantGetInt32(pArg1));

 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}

static void
PrvCmdGetRotation(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand)
{
  //BDBG("Got an GetRotation Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  int retval = 0;
  
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  retval = pCom->GetRotation();

 Done:
  //BDBG("Server returning %d, %d\n", err, retval);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeInt32, retval,
			      kPmIpcTypeNull);
}


static void
PrvCmdGetCurrentPageNumber(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand)
{
  //BDBG("Got an GetCurrentPageNumber Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  uint32_t     retval = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  
  retval = pCom->GetCurrentPageNumber();

 Done:
  //BDBG("Server returning %d, %d\n", err, retval);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeUint32, retval,
			      kPmIpcTypeNull);
}

static void
PrvCmdGetPageCount(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand)
{
  //BDBG("Got an GetPageCount Command!\n");
   PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  uint32_t     retval = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
 
  retval = pCom->GetPageCount();
 Done:
  //BDBG("Server returning %d, %d\n", err, retval);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeUint32, retval,
			      kPmIpcTypeNull);
}

static void
PrvCmdGetPageDimensions(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand)
{
  //BDBG("Got an GetPageDimensions Command!\n");
   PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  int width  = 0;
  int height = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  pCom->GetPageDimensions(width, height);
 Done:
  //BDBG("Server returning %d width: %d, height %d\n", err, width, height);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeUint32, (uint32_t)width,
			      kPmIpcTypeUint32, (uint32_t)height,
			      kPmIpcTypeNull);
}


static void
PrvCmdHasLinkAtLocation(PmIpcClientContext_t* pClientCtxt, void* pUserData, const char* pCommand,
			const PmIpcVariant_t* pArg1, const PmIpcVariant_t* pArg2, 
			const PmIpcVariant_t* pArg3)
{
  //BDBG("Got an GetPageCount Command!\n");
   PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  uint32_t     retval = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  char buffer[2048];
  retval = pCom->HasLinkAtLocation(PmIpcVariantGetUint32(pArg1),PmIpcVariantGetUint32(pArg2), 
				   buffer, PmIpcVariantGetBool(pArg3));
 Done:
  //BDBG("Server returning %d, %d\n", err, retval);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeUint32, retval,
			      kPmIpcTypeString, buffer,
			      kPmIpcTypeNull);
}

static void
PrvCmdOutlineRoot(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		  const char* pCommand)
{
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  pCom->OutlineRoot();
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}

static void
PrvCmdGetOutlineNodeText(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand)
{
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  char*        buffer = NULL;
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  buffer = pCom->GetOutlineNodeText();

 Done:
  //BDBG("Server returning %d, %s\n", err, buffer);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeString, buffer,
			      kPmIpcTypeNull);
}



static void
PrvCmdOutlineGotoChild(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand)
{
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  pCom->OutlineGotoChild();
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}

static void
PrvCmdOutlineHasChild(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand)
{
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  bool         haschild=false;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  haschild = pCom->OutlineHasChild();
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeBool, haschild,
			      kPmIpcTypeNull);
}

static void
PrvCmdOutlineGotoNextSibling(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand)
{
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  pCom->OutlineGotoNextSibling();
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}


static void
PrvCmdOutlineHasNextSibling(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand)
{
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  bool         haschild=false;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  haschild = pCom->OutlineHasNextSibling();
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeBool, haschild,
			      kPmIpcTypeNull);
}

static void
PrvCmdGetOutlineLinkPageNumber(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		      const char* pCommand)
{
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  int         page=0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  page = pCom->GetOutlineLinkPageNumber();
 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeInt32, page,
			      kPmIpcTypeNull);
}



static void
PrvCmdDumpPageText(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
		   const char* pCommand)
{
  //BDBG("Got A DUMP Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  //BDBG("DUMPING!\n\n\n\n\n\n");
  pCom->DumpCurrentPageText();

 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeNull);
}


static void
PrvCmdTranslateScreenPoint(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
			   const char* pCommand, const PmIpcVariant_t* pArg1,
			   const PmIpcVariant_t* pArg2)
{
 //BDBG("Got A Translate POint Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  int x,y=0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  x = PmIpcVariantGetUint32(pArg1);
  y = PmIpcVariantGetUint32(pArg2);
  pCom->TranslateScreenPoint(x,y);

 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeUint32, (uint32_t)x,
			      kPmIpcTypeUint32, (uint32_t)y,
			      kPmIpcTypeNull);

}

static void
PrvCmdTranslateScreenDelta(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
			   const char* pCommand, const PmIpcVariant_t* pArg1,
			   const PmIpcVariant_t* pArg2)
{
  //BDBG("Got A Translate Delta Command!\n");
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
  int x,y=0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  x = PmIpcVariantGetUint32(pArg1);
  y = PmIpcVariantGetUint32(pArg2);
  pCom->TranslateScreenDelta(x,y);

 Done:
  //BDBG("Server returning %d\n", err);
  PmIpcServerSendSyncResponse(pClientCtxt,
			      kPmIpcTypeUint32, err,
			      kPmIpcTypeUint32, (uint32_t)x,
			      kPmIpcTypeUint32, (uint32_t)y,
			      kPmIpcTypeNull);
}

void RecordResult(int pagenumber, int x0, int y0, int x1, int y1, int strlen,char *str, void *userdata)
{
  // BDBG("Server returning %s\n", str);
  PmIpcServerAsyncSend((PmIpcClientContext_t *)userdata, "d.RecordResult",
		       kPmIpcTypeUint32, pagenumber,
		       kPmIpcTypeUint32, x0,
		       kPmIpcTypeUint32, y0,
		       kPmIpcTypeUint32, x1,
		       kPmIpcTypeUint32, y1,
		       kPmIpcTypeString, str,
		       kPmIpcTypeNull);   
}

static void
PrvCmdCaseInsensitiveTextSearch(PmIpcClientContext_t* pClientCtxt, void* pUserData, 
			       const char* pCommand,const PmIpcVariant_t* pArg1,
			       const PmIpcVariant_t* pArg2, const PmIpcVariant_t* pArg3)
{
  PmErr        err   = kPmErrNone;
  DocCommander* pCom = 0;
  void*        pPtr  = 0;
    
  PmIpcServerGetPrivateData(pClientCtxt, &pPtr);
  pCom = (DocCommander*) pPtr;
  if (!pCom) {
    BERR("No Document Loaded for this client: %p", pClientCtxt);
    err = (uint32_t) -10;
    goto Done;
  }
  printf("Got Search Request, looking for %s\n", (char *)PmIpcVariantGetString(pArg1));
  pCom->CaseInsensitiveTextSearch((char *)PmIpcVariantGetString(pArg1), PmIpcVariantGetUint32(pArg2), 
				  PmIpcVariantGetUint32(pArg3), RecordResult, (void *)pClientCtxt);
 Done:
  return;
}
